from __future__ import print_function as _

import os as _os
import sys as _sys
import json

import dash as _dash

# noinspection PyUnresolvedReferences
from ._imports_ import *
from ._imports_ import __all__
from . import utils

# Import CyLeaflet AIO component
from .CyLeaflet import CyLeaflet


if not hasattr(_dash, "__plotly_dash") and not hasattr(_dash, "development"):
    print(
        "Dash was not successfully imported. "
        "Make sure you don't have a file "
        'named \n"dash.py" in your current directory.',
        file=_sys.stderr,
    )
    _sys.exit(1)

_basepath = _os.path.dirname(__file__)
_filepath = _os.path.abspath(_os.path.join(_basepath, "package.json"))
with open(_filepath) as f:
    package = json.load(f)

package_name = package["name"].replace(" ", "_").replace("-", "_")
__version__ = package["version"]

_current_path = _os.path.dirname(_os.path.abspath(__file__))

_this_module = _sys.modules[__name__]


_js_dist = [
    {
        "relative_package_path": "dash_cytoscape.min.js",
        "dev_package_path": "dash_cytoscape.dev.js",
        "external_url": "https://unpkg.com/dash-cytoscape@{2}/{1}/{1}.min.js".format(
            package_name, __name__, __version__
        ),
        "namespace": package_name,
    }
]

_css_dist = []


for _component in __all__:
    setattr(locals()[_component], "_js_dist", _js_dist)
    setattr(locals()[_component], "_css_dist", _css_dist)


def load_extra_layouts():
    """
    Load 3rd party layouts that are not included by default with Cytoscape. You can find the
    documentation about those layouts here:
        - `cose-bilkent`: https://github.com/cytoscape/cytoscape.js-cose-bilkent
        - `cola`: https://github.com/cytoscape/cytoscape.js-cola
        - `euler`: https://github.com/cytoscape/cytoscape.js-dagre
        - `spread`: https://github.com/cytoscape/cytoscape.js-spread
        - `dagre`: https://github.com/cytoscape/cytoscape.js-dagre
        - `klay`: https://github.com/cytoscape/cytoscape.js-klay

    Example:

    ```
    import dash
    from dash import html
    import dash_cytoscape as cyto

    cyto.load_extra_layouts()

    app = dash.Dash(__name__)

    app.layout = html.Div([
        cyto.Cytoscape(...),
    ])
    ```

    Be careful about using the extra layouts when not necessary, since they require supplementary
    bandwidth for loading, which impacts the startup time of the app.
    """
    global _js_dist

    _js_dist = [
        {
            "relative_package_path": "dash_cytoscape_extra.min.js",
            "dev_package_path": "dash_cytoscape_extra.dev.js",
            "external_url": "https://unpkg.com/dash-cytoscape@{}/{}/{}.min.js".format(
                __version__, __name__, "dash_cytoscape_extra"
            ),
            "namespace": package_name,
        }
    ]


def _display_default_values():
    out_string = ""

    metadata_path = _os.path.join(
        _os.path.dirname(_os.path.realpath(__file__)), "metadata.json"
    )

    with open(metadata_path, "r") as file:
        data = json.loads(file.read())

    for component in data:
        component_name = component.replace("src/lib/components/", "").replace(
            ".react.js", ""
        )
        metadata = data[component]
        props = metadata["props"]

        out_string += "## {} Default Values\n\n".format(component_name)

        for prop_name in props:
            prop = props[prop_name]

            if "defaultValue" in prop:
                default = prop["defaultValue"]["value"]

                out_string += "* *{}*: {}\n".format(prop_name, default)

    return out_string



from dash import (
    clientside_callback,
    callback,
    ClientsideFunction,
    Output,
    Input,
    State,
    html,
    dcc,
    MATCH,
    no_update,
)

import dash_cytoscape as cyto
import math


try:
    import dash_leaflet as dl
except ImportError:
    dl = None

# Max zoom of default Leaflet tile layer
LEAFLET_DEFAULT_MAX_ZOOM = 18


# Approximates max zoom values for Cytoscape from max zoom
# values of Leaflet. Empirically-determined
def get_cytoscape_max_zoom(leaflet_max_zoom):
    leaflet_max_zoom = leaflet_max_zoom or 0
    return 0.418 * (2 ** (leaflet_max_zoom - 16))


class CyLeaflet(html.Div):
    # Predefined Leaflet tile layer with max zoom of 19
    OSM = (
        dl.TileLayer(
            url="https://tile.openstreetmap.org/{z}/{x}/{y}.png",
            maxZoom=19,
            attribution='&copy; <a href="https://www.openstreetmap.org/copyright">OpenStreetMap</a> contributors',
        )
        if dl
        else None
    )

    def __init__(
        self,
        id,
        cytoscape_props=None,
        leaflet_props=None,
        width="600px",
        height="480px",
        tiles=None,
    ):
        # Throw error if `dash_leaflet` package is not installed
        if dl is None:
            raise ImportError(
                "dash_leaflet not found. Please install it, either directly (`pip install dash_leaflet`) "
                + "or by using `pip install dash_cytoscape[leaflet]`"
            )

        self.ids = {
            s: {"id": id, "component": "cyleaflet", "sub": s}
            for s in ["cy", "leaf", "elements", "cy-extent"]
        }

        self.CYTOSCAPE_ID = self.ids["cy"]
        self.LEAFLET_ID = self.ids["leaf"]
        self.ELEMENTS_ID = self.ids["elements"]

        cytoscape_props = cytoscape_props or {}
        leaflet_props = leaflet_props or {}
        elements = cytoscape_props.get("elements", [])
        cytoscape_props, leaflet_props = self.set_default_props_and_overrides(
            cytoscape_props, leaflet_props, tiles
        )

        super().__init__(
            html.Div(
                [
                    html.Div(
                        cyto.Cytoscape(**cytoscape_props),
                        style={
                            "height": "100%",
                            "width": "100%",
                            "position": "absolute",
                            "top": 0,
                            "left": 0,
                            "zIndex": 2,
                        },
                    ),
                    html.Div(
                        dl.Map(**leaflet_props),
                        style={
                            "height": "100%",
                            "width": "100%",
                            "position": "absolute",
                            "top": 0,
                            "left": 0,
                            "zIndex": 1,
                        },
                    ),
                    dcc.Store(id=self.ids["elements"], data=elements),
                    dcc.Store(id=self.ids["cy-extent"]),
                ],
                style={
                    "width": width,
                    "height": height,
                },
            ),
            style={
                "height": "100%",
                "width": "100%",
                "position": "relative",
            },
        )

    def set_default_props_and_overrides(
        self, user_cytoscape_props, user_leaflet_props, tiles
    ):
        # If `tiles` is specified, append to end of Leaflet children
        # This will make it the visible TileLayer
        leaflet_children = user_leaflet_props.get("children", [])
        if not isinstance(leaflet_children, list):
            leaflet_children = [leaflet_children]
        if tiles is not None:
            leaflet_children.append(tiles)

        # Try to figure out Leaflet maxZoom from Leaflet children,
        # then convert to Cytoscape max zoom
        leaflet_max_zoom = self.get_leaflet_max_zoom(leaflet_children)
        cytoscape_max_zoom = self.get_cytoscape_max_zoom(leaflet_max_zoom)

        # Props where we want to override values supplied by the user
        # These are props which are required for CyLeaflet to work properly
        cytoscape_overrides = {
            "id": self.ids["cy"],
            "elements": [],  # Elements are set via clientside callback, so set to empty list initially
            "layout": {"name": "preset", "fit": False},
            "style": {"width": "100%", "height": "100%"},
            "minZoom": 3 / 100000,
        }
        # Note: Leaflet MUST be initialized with a center and zoom to avoid an error,
        # even though these will be immediately overwritten by syncing w/ Cytoscape
        leaflet_overrides = {
            "id": self.ids["leaf"],
            "children": leaflet_children or [dl.TileLayer(maxZoom=100000)],
            "center": [0, 0],
            "zoom": 6,
            "zoomSnap": 0,
            "zoomControl": False,
            "zoomAnimation": False,
            "maxZoom": 100000,
            "maxBoundsViscosity": 1,
            "maxBounds": [[-85, -180.0], [85, 180.0]],
            "style": {"width": "100%", "height": "100%"},
        }

        # Props where we want to fill in a default value
        # if a value is not supplied by the user
        cytoscape_defaults = {
            "boxSelectionEnabled": True,
            "maxZoom": cytoscape_max_zoom,
        }
        leaflet_defaults = {}

        # Start with default props
        cytoscape_props = dict(cytoscape_defaults)
        leaflet_props = dict(leaflet_defaults)

        # Update with user-supplied props
        cytoscape_props.update(user_cytoscape_props)
        leaflet_props.update(user_leaflet_props)

        # Update with overrides
        cytoscape_props.update(cytoscape_overrides)
        leaflet_props.update(leaflet_overrides)

        return cytoscape_props, leaflet_props

    # Try to figure out Leaflet maxZoom from Leaflet children
    # If not possible, return the maxZoom of the default Leaflet tile layer
    def get_leaflet_max_zoom(self, leaflet_children):
        if leaflet_children is None or leaflet_children == []:
            return LEAFLET_DEFAULT_MAX_ZOOM

        max_zooms = [
            c.maxZoom
            for c in leaflet_children
            if isinstance(c, dl.TileLayer) and hasattr(c, "maxZoom")
        ]

        return max_zooms[-1] if len(max_zooms) > 0 else LEAFLET_DEFAULT_MAX_ZOOM

    # Given a maxZoom value for Leaflet, map it to the corresponding maxZoom value for Cytoscape
    # If the value is out of range, return the closest value
    def get_cytoscape_max_zoom(self, leaflet_max_zoom):
        return get_cytoscape_max_zoom(leaflet_max_zoom)


if dl is not None:
    clientside_callback(
        ClientsideFunction(namespace="cyleaflet", function_name="updateLeafBounds"),
        Output(
            {"id": MATCH, "component": "cyleaflet", "sub": "leaf"}, "invalidateSize"
        ),
        Output({"id": MATCH, "component": "cyleaflet", "sub": "leaf"}, "viewport"),
        Output({"id": MATCH, "component": "cyleaflet", "sub": "cy-extent"}, "data"),
        Input({"id": MATCH, "component": "cyleaflet", "sub": "cy"}, "extent"),
        Input({"id": MATCH, "component": "cyleaflet", "sub": "cy"}, "maxZoom"),
        State({"id": MATCH, "component": "cyleaflet", "sub": "cy-extent"}, "data"),
    )
    clientside_callback(
        ClientsideFunction(namespace="cyleaflet", function_name="transformElements"),
        Output(
            {"id": MATCH, "component": "cyleaflet", "sub": "cy"},
            "elements",
            allow_duplicate=True,
        ),
        Input({"id": MATCH, "component": "cyleaflet", "sub": "elements"}, "data"),
        prevent_initial_call="initial_duplicate",
    )
    clientside_callback(
        ClientsideFunction(namespace="cyleaflet", function_name="updateLonLat"),
        Output({"id": MATCH, "component": "cyleaflet", "sub": "elements"}, "data"),
        Input({"id": MATCH, "component": "cyleaflet", "sub": "cy"}, "elements"),
        prevent_initial_call=True,
    )
    clientside_callback(
        ClientsideFunction(namespace="cyleaflet", function_name="updateCytoMaxZoom"),
        Output({"id": MATCH, "component": "cyleaflet", "sub": "cy"}, "maxZoom"),
        Input({"id": MATCH, "component": "cyleaflet", "sub": "leaf"}, "children"),
        prevent_initial_call=True,
    )



"""Notes for the hardcoded values in the `init_pos` dictionary:

It's impossible to programatically get the initial rendered positions of the nodes, since we would
need to obtain the node object (in dictionary format), which can be done either by (a) using the
tapNode callback, which paradoxally requires you to click the node, or (b) by making a complex
calculation relative to the size of the screen w.r.t largest coordinate in the list of elements.
But (b) is unreliable, since we do not know how much padding around the graph is required, so it
will likely be off.

If there is a need to modify the values in `init_pos`, e.g. if the size of the webdriver screen
is changed, you can do the following:
    - Run usage-events.py
    - Resize window size to 1280x1000, or preferred size (can be manually done or with selenium)
    - Tap on a node
    - Inside the "Node Object JSON" section, find "renderedPosition" and use the values there
    - Repeat this for all the nodes

Notice also that there's an offset to Node 3's position. This is because it overlaps with
Node 6, so clicking on Node 3 will erroneously show that you clicked Node 6. Therefore, adding an
offset to the y-axis will ensure that the correct node is clicked.
"""
import os
import importlib
import time
import json

from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.common.by import By


def create_app(dash_duo):
    # Initialize the apps
    app = importlib.import_module("usage-events").app

    dash_duo.start_server(app)
    dash_duo.driver.set_window_size(1280, 1000)
    dash_duo.wait_for_element_by_id("cytoscape", 20)

    actions = ActionChains(dash_duo.driver)
    init_pos = {
        "Node 1": (59, 182),
        "Node 2": (222, 345),
        "Node 3": (168, 283 - 20),
        "Node 4": (440, 182),
        "Node 5": (277, 236),
        "Node 6": (168, 283),
    }

    def calculate_edge_position(node_1, node_2):
        x = round((node_1[0] + node_2[0]) * 0.5)
        y = round((node_1[1] + node_2[1]) * 0.5)
        return (x, y)

    edges_positions = {
        "Edge from Node 1 to Node 2": calculate_edge_position(
            init_pos["Node 1"], init_pos["Node 2"]
        ),
        "Edge from Node 2 to Node 4": calculate_edge_position(
            init_pos["Node 2"], init_pos["Node 4"]
        ),
        "Edge from Node 5 to Node 1": calculate_edge_position(
            init_pos["Node 5"], init_pos["Node 1"]
        ),
    }
    return init_pos, actions, edges_positions


def save_screenshot(dash_duo, dir_name, name):
    directory_path = os.path.join(os.path.dirname(__file__), "screenshots", dir_name)

    # Create directory if it doesn't already exist
    if not os.path.exists(directory_path):
        os.makedirs(directory_path)

    dash_duo.driver.save_screenshot(
        os.path.join(os.path.dirname(__file__), "screenshots", dir_name, name + ".png")
    )


def perform_dragging(
    dash_duo, x, y, delta_x, delta_y, elem, actions, dir_name="interactions"
):
    """
    Performs dragging on a node, and return the difference from the start
    :param x: initial position of the node at the start of the action chain
    :param y: initial position of the node at the start of the action chain
    :param delta_x: how much we want to drag the node
    :param delta_y: how much we want to drag the node
    :param dir_name: The directory in which we store our screenshots
    :return: the difference between the position after drag and the starting position
    """
    actions.reset_actions()
    actions.move_to_element_with_offset(
        dash_duo.driver.find_element(By.TAG_NAME, "body"), x, y
    )
    actions.drag_and_drop_by_offset(source=None, xoffset=delta_x, yoffset=delta_y)
    actions.perform()
    time.sleep(1)

    elem_json = json.loads(elem.text)
    new_pos = (
        elem_json[0].get("position")
        if "position" in elem_json[0].keys()
        else elem_json[0].get("renderedPosition")
    )
    dragged_label = elem_json[0].get("data", {}).get("label")

    node_x = round(new_pos["x"])
    node_y = round(new_pos["y"])

    save_screenshot(
        dash_duo,
        dir_name,
        f"Dragged{dragged_label.replace(' ', '')}By{node_x}x{node_y}y",
    )

    return node_x, node_y


def perform_clicking(dash_duo, x, y, elem, actions, dir_name="interactions"):
    """
    :param x: The position on the screen where we want to click
    :param y: The position on the screen where we want to click
    :param elem: The element object from where we retrieve the JSON
    :param dir_name: The directory in which we store our screenshots
    :return: The label of element most recently clicked, if any
    """
    actions.reset_actions()
    actions.move_to_element_with_offset(
        dash_duo.driver.find_element(By.TAG_NAME, "body"), x, y
    )
    actions.click()
    actions.perform()

    time.sleep(1)
    clicked_label = json.loads(elem.text).get("data", {}).get("label")

    save_screenshot(dash_duo, dir_name, "Clicked" + clicked_label.replace(" ", ""))

    return clicked_label


def perform_mouseover(
    dash_duo, x, y, elem, actions, dir_name="interactions", screenshot_name="Mouseover"
):
    actions.reset_actions()
    actions.move_to_element_with_offset(
        dash_duo.driver.find_element(By.TAG_NAME, "body"), x - 50, y
    )
    actions.move_by_offset(50, 0)
    actions.perform()
    time.sleep(1)

    text = json.loads(elem.text)
    mouseover_label = json.loads(elem.text).get("label") if text else "null"

    save_screenshot(
        dash_duo, dir_name, screenshot_name + mouseover_label.replace(" ", "")
    )

    return mouseover_label


def test_cyin001_dragging(dash_duo):
    init_pos, actions, _ = create_app(dash_duo)

    # View module docstring for more information about initial positions
    init_x, init_y = init_pos["Node 1"]

    # Open the Drag data JSON tab
    actions.move_to_element(dash_duo.find_element("#tabs > div:nth-child(5)"))
    actions.click().perform()
    time.sleep(1)

    # Select the JSON output element
    elem_tap = dash_duo.find_element("pre#elements-data-json-output")

    # Get initial positions. Not actualy dragging
    init_node_x, init_node_y = perform_dragging(
        dash_duo, init_x, init_y, 1, 1, elem_tap, actions
    )

    def _test(newpos, expected_shift):
        dx, dy = expected_shift
        newx, newy = newpos
        expx = (dx / pixels_to_position_conv_factor) + init_node_x
        expy = (dy / pixels_to_position_conv_factor) + init_node_y
        assert abs(newx - expx) < 3
        assert abs(newy - expy) < 3

    pixels_to_position_conv_factor = 1280 * 0.00085
    # Test dragging the nodes around
    _test(
        perform_dragging(dash_duo, init_x, init_y, 150, 0, elem_tap, actions), (150, 0)
    )
    _test(
        perform_dragging(dash_duo, init_x + 150, init_y, 0, 150, elem_tap, actions),
        (150, 150),
    )
    _test(
        perform_dragging(
            dash_duo, init_x + 150, init_y + 150, -150, -150, elem_tap, actions
        ),
        (0, 0),
    )
    _test(
        perform_dragging(dash_duo, init_x, init_y, 100, -100, elem_tap, actions),
        (100, -100),
    )


def test_cyin002_clicking(dash_duo):
    init_pos, actions, _ = create_app(dash_duo)
    click_error = "Unable to click Cytoscape nodes properly"

    # Select the JSON output element
    elem_tap = dash_duo.find_element("pre#tap-node-json-output")

    # Test clicking the nodes
    for i in range(1, 7):
        label = f"Node {i}"
        assert (
            perform_clicking(dash_duo, *init_pos[label], elem_tap, actions) == label
        ), click_error


def test_cyin003_clicking_edges(dash_duo):
    _, actions, edges_positions = create_app(dash_duo)

    # Select the JSON output element
    elem_tap = dash_duo.find_element("pre#tap-edge-json-output")

    # Test clicking the edges
    for label, positions in edges_positions.items():
        assert perform_clicking(dash_duo, *positions, elem_tap, actions) == label


def test_cyin004_mouseover(dash_duo):
    init_pos, actions, _ = create_app(dash_duo)
    mouseover_error = "Unable to mouseover Cytoscape nodes properly"

    # Open the Mouseover JSON tab
    actions.move_to_element(dash_duo.find_element("#tabs > div:nth-child(3)"))
    actions.click().perform()
    time.sleep(1)

    # Select the JSON output element
    elem_mouseover = dash_duo.find_element("pre#mouseover-node-data-json-output")

    # Test hovering the nodes
    for i in range(1, 7):
        label = f"Node {i}"
        assert (
            perform_mouseover(dash_duo, *init_pos[label], elem_mouseover, actions)
            == label
        ), mouseover_error


def test_cyin005_mouseover_edges(dash_duo):
    _, actions, edges_positions = create_app(dash_duo)

    # Open the Mouseover JSON tab
    actions.move_to_element(dash_duo.find_element("#tabs > div:nth-child(3)"))
    actions.click().perform()
    time.sleep(1)

    # Select the JSON output element
    elem_mouseover = dash_duo.find_element("pre#mouseover-edge-data-json-output")

    # Test mouseover the edges
    for label, positions in edges_positions.items():
        assert perform_mouseover(dash_duo, *positions, elem_mouseover, actions) == label


def test_cyin006_mouseover_unhover(dash_duo):
    init_pos, actions, _ = create_app(dash_duo)
    mouseover_error = "Unable to mouseover Cytoscape nodes properly"

    # Open the Mouseover JSON tab
    actions.move_to_element(dash_duo.find_element("#tabs > div:nth-child(3)"))
    actions.click().perform()
    time.sleep(1)

    # Select the JSON output element
    elem_mouseover = dash_duo.find_element("pre#mouseover-node-data-json-output")

    # Test hovering the nodes
    label = "Node 1"
    assert (
        perform_mouseover(
            dash_duo,
            *init_pos[label],
            elem_mouseover,
            actions,
            screenshot_name="MouseoverHover",
        )
        == label
    ), mouseover_error

    # Test unhover
    label = "null"
    assert (
        perform_mouseover(
            dash_duo,
            70,
            250,
            elem_mouseover,
            actions,
            screenshot_name="MouseoverUnhover",
        )
        == label
    ), mouseover_error


def test_cyin007_click_twice(dash_duo):
    init_pos, actions, _ = create_app(dash_duo)

    # Open the Tap Data JSON tab
    actions.move_to_element(dash_duo.find_element("#tabs > div:nth-child(2)"))
    actions.click().perform()
    time.sleep(1)

    # Select the JSON output element
    elem_tap = dash_duo.find_element("pre#tap-node-data-json-output")

    # Test clicking the same node twice
    label = "Node 1"
    x, y = init_pos[label]

    actions.reset_actions()
    actions.move_to_element_with_offset(
        dash_duo.driver.find_element(By.TAG_NAME, "body"), x, y
    )
    actions.click()
    actions.perform()

    time.sleep(1)
    clicked_label_1 = json.loads(elem_tap.text).get("label")
    clicked_timestamp_1 = json.loads(elem_tap.text).get("timeStamp")

    # Second click
    actions.click()
    actions.perform()

    time.sleep(1)
    clicked_label_2 = json.loads(elem_tap.text).get("label")
    clicked_timestamp_2 = json.loads(elem_tap.text).get("timeStamp")

    assert clicked_label_1 == clicked_label_2
    assert clicked_timestamp_1 != clicked_timestamp_2


def test_cyin008_ctx_menu_remove_node(dash_duo):
    _, actions, _ = create_app(dash_duo)

    # Open the Drag data JSON tab
    actions.move_to_element(dash_duo.find_element("#tabs > div:nth-child(5)"))
    actions.click().perform()
    time.sleep(1)

    # Select the JSON output element before removal
    elements = dash_duo.find_element("pre#elements-data-json-output")
    nb_elements_before = len(json.loads(elements.text))

    # move mouse to first node and right click and click on remove
    actions.move_by_offset(-1170, 130)
    actions.context_click()
    actions.move_to_element(dash_duo.find_element("button#remove"))
    actions.click()
    actions.perform()
    time.sleep(1)

    # Select the JSON output element after removal
    elements = dash_duo.find_element("pre#elements-data-json-output")
    nb_elements_after = len(json.loads(elements.text))
    # removed one node and 3 edges
    assert nb_elements_before - 4 == nb_elements_after


def test_cyin009_ctx_menu_remove_edge(dash_duo):
    _, actions, _ = create_app(dash_duo)

    # Open the Drag data JSON tab
    actions.move_to_element(dash_duo.find_element("#tabs > div:nth-child(5)"))
    actions.click().perform()
    time.sleep(1)

    # Select the JSON output element before removal
    elements = dash_duo.find_element("pre#elements-data-json-output")
    nb_elements_before = len(json.loads(elements.text))

    # move mouse to an edge and right click and click on remove
    actions.move_by_offset(-1100, 150)
    actions.context_click()
    actions.move_to_element(dash_duo.find_element("button#remove"))
    actions.click()
    actions.perform()
    time.sleep(1)

    # Select the JSON output element after removal
    elements = dash_duo.find_element("pre#elements-data-json-output")
    nb_elements_after = len(json.loads(elements.text))

    assert nb_elements_before - 1 == nb_elements_after


def test_cyin010_ctx_menu_add_node(dash_duo):
    _, actions, _ = create_app(dash_duo)

    # Open the Drag data JSON tab
    actions.move_to_element(dash_duo.find_element("#tabs > div:nth-child(5)"))
    actions.click().perform()
    time.sleep(1)

    # Select the JSON output element before addition
    elements = dash_duo.find_element("pre#elements-data-json-output")
    nb_elements_before = len(json.loads(elements.text))

    # click anywhere to add a node
    actions.move_by_offset(-1000, 40)
    actions.context_click()
    actions.move_to_element(dash_duo.find_element("button#add-node"))
    actions.click()
    actions.perform()
    time.sleep(1)

    # Select the JSON output element after addition
    elements = dash_duo.find_element("pre#elements-data-json-output")
    nb_elements_after = len(json.loads(elements.text))

    assert nb_elements_before + 1 == nb_elements_after


def test_cyin011_ctx_menu_add_edge(dash_duo):
    _, actions, _ = create_app(dash_duo)

    # Open the Drag data JSON tab
    actions.move_to_element(dash_duo.find_element("#tabs > div:nth-child(5)"))
    actions.click().perform()
    time.sleep(1)

    # Select the JSON output element before addition
    elements = dash_duo.find_element("pre#elements-data-json-output")
    nb_elements_before = len(json.loads(elements.text))

    # click on 2 nodes to add an edge between them
    actions.key_down(Keys.COMMAND)
    actions.move_by_offset(-1170, 130)
    actions.click()
    actions.move_by_offset(400, 0)
    actions.click()
    actions.context_click()
    actions.move_to_element(dash_duo.find_element("button#add-edge"))
    actions.click()
    actions.perform()
    time.sleep(1)

    # Select the JSON output element after addition
    elements = dash_duo.find_element("pre#elements-data-json-output")
    nb_elements_after = len(json.loads(elements.text))

    assert nb_elements_before + 1 == nb_elements_after



"""
In order to render snapshots, Percy collects the DOM of the project and
uses a custom rendering method, different from Selenium. Therefore, it
is unable to render Canvas elements, so can't render Cytoscape charts
directly.

Instead, we use Selenium webdrivers to automatically screenshot each of
the apps being tested in test_usage.py, display them in a simple
Dash app, and use Percy to take a snapshot for CVI.

Here, we extend the setUpClass method from IntegrationTests by adding
percy runner initialization. This is because other classes that inherits
from IntegrationTests do not necessarily need to initialize Percy (since
all they do is save snapshots), and doing so causes Percy to render an
empty build that ends up failing. Therefore, we decide to initialize and
finalize the Percy runner in this class rather than inside
IntegrationTests.
"""

import base64
import os
import sys
import time
import pytest
import dash
from dash import html, dcc, Input, Output, callback


def create_app(dir_name):
    def encode(name):
        path = os.path.join(os.path.dirname(__file__), "screenshots", dir_name, name)

        with open(path, "rb") as image_file:
            encoded_string = base64.b64encode(image_file.read())
        return "data:image/png;base64," + encoded_string.decode("ascii")

    # Define the app
    app = dash.Dash(__name__)

    app.layout = html.Div(
        [
            # represents the URL bar, doesn't render anything
            dcc.Location(id="url", refresh=False),
            # content will be rendered in this element
            html.Div(id="page-content"),
        ]
    )

    @callback(
        Output("page-content", "children"),
        Input("url", "pathname"),
    )
    def display_image(pathname):  # pylint: disable=W0612
        """
        Assign the url path to return the image it represent. For example,
        to return "usage.png", you can visit localhost/usage.png.
        :param pathname: name of the screenshot, prefixed with "/"
        :return: An html.Img object containing the base64 encoded image
        """
        if not pathname or pathname == "/":
            return None

        name = pathname.replace("/", "")
        return html.Img(id=name, src=encode(name))

    return app


def percy_snapshot(dash_duo, name=""):
    snapshot_name = f"{name} (Python {sys.version_info.major}.{sys.version_info.minor})"

    dash_duo.percy_snapshot(name=snapshot_name)


def run_percy_on(dir_name, dash_duo):
    # Find the names of all the screenshots
    asset_list = os.listdir(
        os.path.join(os.path.dirname(__file__), "screenshots", dir_name)
    )

    current_url = dash_duo.driver.current_url
    # Run Percy
    for image in asset_list:
        if image.endswith("png"):
            output_name = image.replace(".png", "")

            dash_duo.driver.get(current_url + image)

            dash_duo.wait_for_element_by_id(image, 20)

            percy_snapshot(dash_duo, name=f"{dir_name.upper()}: {output_name}")
            time.sleep(2)


@pytest.mark.parametrize(
    "name", ["usage", "elements", "layouts", "style", "interactions"]
)
def test_cyps001_snapshots(name, dash_duo):
    app = create_app(dir_name=name)
    dash_duo.start_server(app)

    run_percy_on(name, dash_duo)



# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class Cytoscape(Component):
    """A Cytoscape component.
    A Component Library for Dash aimed at facilitating network visualization in
    Python, wrapped around [Cytoscape.js](http://js.cytoscape.org/).

    Keyword arguments:

    - id (string; optional):
        The ID used to identify this component in Dash callbacks.

    - autoRefreshLayout (boolean; default True):
        Whether the layout should be refreshed when elements are added or
        removed.

    - autolock (boolean; default False):
        Whether nodes should be locked (not draggable at all) by default
        (if True, overrides individual node state).

    - autoungrabify (boolean; default False):
        Whether nodes should be ungrabified (not grabbable by user) by
        default (if True, overrides individual node state).

    - autounselectify (boolean; default False):
        Whether nodes should be unselectified (immutable selection state)
        by default (if True, overrides individual element state).

    - boxSelectionEnabled (boolean; default False):
        Whether box selection (i.e. drag a box overlay around, and release
        it to select) is enabled. If enabled, the user must taphold to pan
        the graph.

    - className (string; optional):
        Sets the class name of the element (the value of an element's html
        class attribute).

    - clearOnUnhover (boolean; default False):
        If set to True, mouseoverNodeData and mouseoverEdgeData will be
        cleared on unhover. If set to False, the value of
        mouseoverNodeData and mouseoverEdgeData will be the last Node or
        Edge hovered over.

    - contextMenu (list of dicts; optional):
        Define a custom context menu. The behaviour of each menu item can
        be defined in 1 of 3 ways. 1. By passing a string to onClick that
        refers to one of the built-in Javascript functions. 2. By passing
        a string to onClickCustom that refers to one of the user-defined
        functions in a namespace. 3. By omitting both of these properties;
        this will update the contextMenuData property and trigger a Dash
        callback.

        `contextMenu` is a list of dicts with keys:

        - availableOn (list; optional):
            A list containing either 'node', 'edge',and/or 'canvas'. This
            will determine where the context  menu item will show up.

        - id (string; required):
            ID of the menu item in the context menu.

        - label (string; required):
            The label on the context menu item.

        - onClick (string; optional):
            Specify which built-in JavaScript function to use as behaviour
            for the context menu item. One of 'remove', 'add_node', or
            'add_edge'.

        - onClickCustom (string; optional):
            Specify which user-defined Javascript function to use in the
            dashCytoscapeFunctions namespace as behaviour for the context
            menu item.

        - tooltipText (string; optional):
            The tooltip text when hovering on top of a context menu item.

    - contextMenuData (dict; optional):
        Retrieve relevant data when a context menu item is clicked.
        Read-only.

        `contextMenuData` is a dict with keys:

        - edgeSource (string; optional):
            Node ID of the edge source if the clicked element is an edge,
            or else this property is not returned.

        - edgeTarget (string; optional):
            Node ID of the edge target if the clicked element is an edge,
            or else this property is not returned.

        - elementId (string; optional):
            Element ID on context click if the context click was on an
            element. If context click was on white space, this property is
            not returned.

        - menuItemId (string; optional):
            ID of the menu item in the context menu.

        - timeStamp (number; optional):
            Timestamp of context click.

        - x (number; optional):
            x-position of the context click.

        - y (number; optional):
            y-position of the context click.

    - elements (list of dicts; optional):
        A list of dictionaries representing the elements of the networks.
        Each dictionary describes an element, and specifies its purpose.
        The [official Cytoscape.js
        documentation](https://js.cytoscape.org/#notation/elements-json)
        offers an extensive overview and examples of element declaration.
        Alternatively, a dictionary with the format { 'nodes': [],
        'edges': [] } is allowed at initialization, but arrays remain the
        recommended format.

        `elements` is a list of dicts with keys:

        - classes (string; optional):
            Space separated string of class names of the element. Those
            classes can be selected by a style selector.

        - data (dict; optional):
            Element specific data.

            `data` is a dict with keys:

            - id (string; optional):
                Reference to the element, useful for selectors and edges.
                Randomly assigned if not given.

            - label (string; optional):
                Optional name for the element, useful when `data(label)`
                is given to a style's `content` or `label`. It is only a
                convention.

            - parent (string; optional):
                Only for nodes. Optional reference to another node. Needed
                to create compound nodes.

            - source (string; optional):
                Only for edges. The id of the source node, which is where
                the edge starts.

            - target (string; optional):
                Only for edges. The id of the target node, where the edge
                ends.

        - grabbable (boolean; optional):
            Only for nodes. If the node can be grabbed and moved by the
            user.

        - group (string; optional):
            Either 'nodes' or 'edges'. If not given, it's automatically
            inferred.

        - locked (boolean; optional):
            Only for nodes. If the position is immutable.

        - position (dict; optional):
            Only for nodes. The position of the node.

            `position` is a dict with keys:

            - x (number; optional):
                The x-coordinate of the node.

            - y (number; optional):
                The y-coordinate of the node.

        - selectable (boolean; optional):
            If the element can be selected.

        - selected (boolean; optional):
            If the element is selected upon initialisation.

          Or dict with keys:

        - edges (list; optional)

        - nodes (list; optional)

    - extent (dict; optional):
        Extent of the viewport, a bounding box in model co-ordinates that
        lets you know what model positions are visible in the viewport.
        This function returns a plain object bounding box with format {
        x1, y1, x2, y2, w, h }.

    - generateImage (dict; optional):
        Dictionary specifying options to generate an image of the current
        cytoscape graph. Value is cleared after data is received and image
        is generated. This property will be ignored on the initial
        creation of the cytoscape object and must be invoked through a
        callback after it has been rendered.  If the app does not need the
        image data server side and/or it will only be used to download the
        image, it may be prudent to invoke `'download'` for `action`
        instead of `'store'` to improve performance by preventing transfer
        of data to the server.

        `generateImage` is a dict with keys:

        - action (a value equal to: 'store', 'download', 'both'; optional):
            `'store'`: Stores the image data (only jpg and png are
            supported) in `imageData` and invokes server-side Dash
            callbacks. `'download'`: Downloads the image as a file with
            all data handling done client-side. No `imageData` callbacks
            are fired. `'both'`: Stores image data and downloads image as
            file. The default is `'store'`.

        - filename (string; optional):
            Name for the file to be downloaded. Default: 'cyto'.

        - options (dict; optional):
            Dictionary of options to cy.png() / cy.jpg() or cy.svg() for
            image generation. See https://js.cytoscape.org/#core/export
            for details. For `'output'`, only 'base64' and 'base64uri' are
            supported. Default: `{'output': 'base64uri'}`.

        - type (a value equal to: 'svg', 'png', 'jpg', 'jpeg'; optional):
            File type to output.

    - imageData (string; optional):
        String representation of the image requested with generateImage.
        Null if no image was requested yet or the previous request failed.
        Read-only.

    - layout (dict; default {name: 'grid'}):
        A dictionary specifying how to set the position of the elements in
        your graph. The `'name'` key is required, and indicates which
        layout (algorithm) to use. The keys accepted by `layout` vary
        depending on the algorithm, but these keys are accepted by all
        layouts: `fit`,  `padding`, `animate`, `animationDuration`,
        `boundingBox`.   The complete list of layouts and their accepted
        options are available on the  [Cytoscape.js
        docs](https://js.cytoscape.org/#layouts) . For the external
        layouts, the options are listed in the \"API\" section of the
        README.  Note that certain keys are not supported in Dash since
        the value is a JavaScript  function or a callback. Please visit
        this [issue](https://github.com/plotly/dash-cytoscape/issues/25)
        for more information.

        `layout` is a dict with keys:

        - animate (boolean; optional):
            Whether to animate change in position when the layout changes.

        - animationDuration (number; optional):
            Duration of animation in milliseconds, if enabled.

        - boundingBox (dict; optional):
            How to constrain the layout in a specific area. Keys accepted
            are either `x1, y1, x2, y2` or `x1, y1, w, h`, all of which
            receive a pixel value.

        - fit (boolean; optional):
            Whether to render the nodes in order to fit the canvas.

        - name (a value equal to: 'random', 'preset', 'circle', 'concentric', 'grid', 'breadthfirst', 'cose', 'cose-bilkent', 'fcose', 'cola', 'euler', 'spread', 'dagre', 'klay'; required):
            The layouts available by default are:   `random`: Randomly
            assigns positions.   `preset`: Assigns position based on the
            `position` key in element dictionaries.   `circle`:
            Single-level circle, with optional radius.   `concentric`:
            Multi-level circle, with optional radius.   `grid`: Square
            grid, optionally with numbers of `rows` and `cols`.
            `breadthfirst`: Tree structure built using BFS, with optional
            `roots`.   `cose`: Force-directed physics simulation.  Some
            external layouts are also included. To use them, run
            `dash_cytoscape.load_extra_layouts()` before creating your
            Dash app. Be careful about   using the extra layouts when not
            necessary, since they require supplementary bandwidth   for
            loading, which impacts the startup time of the app.   The
            external layouts are:
            [cose-bilkent](https://github.com/cytoscape/cytoscape.js-cose-bilkent),
            [fcose](https://github.com/iVis-at-Bilkent/cytoscape.js-fcose),
            [cola](https://github.com/cytoscape/cytoscape.js-cola),
            [euler](https://github.com/cytoscape/cytoscape.js-dagre),
            [spread](https://github.com/cytoscape/cytoscape.js-spread),
            [dagre](https://github.com/cytoscape/cytoscape.js-dagre),
            [klay](https://github.com/cytoscape/cytoscape.js-klay),.

        - padding (number; optional):
            Padding around the sides of the canvas, if fit is enabled.

    - maxZoom (number; default 1e50):
        A maximum bound on the zoom level of the graph. The viewport can
        not be scaled larger than this zoom level.

    - minZoom (number; default 1e-50):
        A minimum bound on the zoom level of the graph. The viewport can
        not be scaled smaller than this zoom level.

    - mouseoverEdgeData (dict; optional):
        The data dictionary of an edge returned when you hover over it.
        Read-only.

    - mouseoverNodeData (dict; optional):
        The data dictionary of a node returned when you hover over it.
        Read-only.

    - pan (dict; default {x: 0, y: 0}):
        Dictionary indicating the initial panning position of the graph.
        The following keys are accepted:.

        `pan` is a dict with keys:

        - x (number; optional):
            The x-coordinate of the node.

        - y (number; optional):
            The y-coordinate of the node.

    - panningEnabled (boolean; default True):
        Whether panning the graph is enabled (i.e., the position of the
        graph is mutable overall).

    - responsive (boolean; default False):
        Toggles intelligent responsive resize of Cytoscape graph with
        viewport size change.

    - selectedEdgeData (list; optional):
        The list of data dictionaries of all selected edges (e.g. using
        Shift+Click to select multiple nodes, or Shift+Drag to use box
        selection). Read-only.

    - selectedNodeData (list; optional):
        The list of data dictionaries of all selected nodes (e.g. using
        Shift+Click to select multiple nodes, or Shift+Drag to use box
        selection). Read-only.

    - style (dict; default {width: '600px', height: '600px'}):
        Add inline styles to the root element.

    - stylesheet (list of dicts; optional):
        A list of dictionaries representing the styles of the elements.
        Each dictionary requires the following keys: `selector` and
        `style`.  Both the [selector](https://js.cytoscape.org/#selectors)
        and the [style](https://js.cytoscape.org/#style/node-body) are
        exhaustively documented in the Cytoscape.js docs. Although methods
        such as `cy.elements(...)` and `cy.filter(...)` are not available,
        the selector string syntax stays the same.

        `stylesheet` is a list of dicts with keys:

        - selector (string; required):
            Which elements you are styling. Generally, you select a group
            of elements (node, edges, both), a class (that you declare in
            the element dictionary), or an element by ID.

        - style (dict; required):
            What aspects of the elements you want to modify. This could be
            the size or color of a node, the shape of an edge arrow, or
            many more.

    - tapEdge (dict; optional):
        The complete edge dictionary returned when you tap or click it.
        Read-only.

        `tapEdge` is a dict with keys:

        - classes (string; optional):
            General item (for all elements).

        - data (dict; optional):
            General item (for all elements).

        - grabbable (boolean; optional):
            General item (for all elements).

        - group (string; optional):
            General item (for all elements).

        - isLoop (boolean; optional):
            Edge-specific item.

        - isSimple (boolean; optional):
            Edge-specific item.

        - locked (boolean; optional):
            General item (for all elements).

        - midpoint (dict; optional):
            Edge-specific item.

        - selectable (boolean; optional):
            General item (for all elements).

        - selected (boolean; optional):
            General item (for all elements).

        - sourceData (dict; optional):
            Edge-specific item.

        - sourceEndpoint (dict; optional):
            Edge-specific item.

        - style (dict; optional):
            General item (for all elements).

        - targetData (dict; optional):
            Edge-specific item.

        - targetEndpoint (dict; optional):
            Edge-specific item.

        - timeStamp (number; optional):
            Edge-specific item.

    - tapEdgeData (dict; optional):
        The data dictionary of an edge returned when you tap or click it.
        Read-only.

    - tapNode (dict; optional):
        The complete node dictionary returned when you tap or click it.
        Read-only.

        `tapNode` is a dict with keys:

        - ancestorsData (dict | list; optional):
            Item for compound nodes.

        - childrenData (dict | list; optional):
            Item for compound nodes.

        - classes (string; optional):
            General item (for all elements).

        - data (dict; optional):
            General item (for all elements).

        - descendantsData (dict | list; optional):
            Item for compound nodes.

        - edgesData (list; optional):
            node specific item.

        - grabbable (boolean; optional):
            General item (for all elements).

        - group (string; optional):
            General item (for all elements).

        - isChild (boolean; optional):
            Item for compound nodes.

        - isChildless (boolean; optional):
            Item for compound nodes.

        - isOrphan (boolean; optional):
            Item for compound nodes.

        - isParent (boolean; optional):
            Item for compound nodes.

        - locked (boolean; optional):
            General item (for all elements).

        - parentData (dict | list; optional):
            Item for compound nodes.

        - position (dict; optional):
            General item (for all elements).

        - relativePosition (dict; optional):
            Item for compound nodes.

        - renderedPosition (dict; optional):
            node specific item.

        - selectable (boolean; optional):
            General item (for all elements).

        - selected (boolean; optional):
            General item (for all elements).

        - siblingsData (dict | list; optional):
            Item for compound nodes.

        - style (dict; optional):
            General item (for all elements).

        - timeStamp (number; optional):
            node specific item.

    - tapNodeData (dict; optional):
        The data dictionary of a node returned when you tap or click it.
        Read-only.

    - userPanningEnabled (boolean; default True):
        Whether user events (e.g. dragging the graph background) are
        allowed to pan the graph.

    - userZoomingEnabled (boolean; default True):
        Whether user events (e.g. dragging the graph background) are
        allowed to pan the graph.

    - wheelSensitivity (number; default 1):
        Changes the scroll wheel sensitivity when zooming.

    - zoom (number; default 1):
        The initial zoom level of the graph. You can set `minZoom` and
        `maxZoom` to set restrictions on the zoom level.

    - zoomingEnabled (boolean; default True):
        Whether zooming the graph is enabled (i.e., the zoom level of the
        graph is mutable overall)."""

    _children_props = []
    _base_nodes = ["children"]
    _namespace = "dash_cytoscape"
    _type = "Cytoscape"

    @_explicitize_args
    def __init__(
        self,
        id=Component.UNDEFINED,
        className=Component.UNDEFINED,
        style=Component.UNDEFINED,
        elements=Component.UNDEFINED,
        stylesheet=Component.UNDEFINED,
        layout=Component.UNDEFINED,
        contextMenu=Component.UNDEFINED,
        contextMenuData=Component.UNDEFINED,
        pan=Component.UNDEFINED,
        zoom=Component.UNDEFINED,
        panningEnabled=Component.UNDEFINED,
        userPanningEnabled=Component.UNDEFINED,
        minZoom=Component.UNDEFINED,
        maxZoom=Component.UNDEFINED,
        zoomingEnabled=Component.UNDEFINED,
        userZoomingEnabled=Component.UNDEFINED,
        wheelSensitivity=Component.UNDEFINED,
        boxSelectionEnabled=Component.UNDEFINED,
        autoungrabify=Component.UNDEFINED,
        autolock=Component.UNDEFINED,
        autounselectify=Component.UNDEFINED,
        autoRefreshLayout=Component.UNDEFINED,
        tapNode=Component.UNDEFINED,
        tapNodeData=Component.UNDEFINED,
        tapEdge=Component.UNDEFINED,
        tapEdgeData=Component.UNDEFINED,
        mouseoverNodeData=Component.UNDEFINED,
        mouseoverEdgeData=Component.UNDEFINED,
        selectedNodeData=Component.UNDEFINED,
        selectedEdgeData=Component.UNDEFINED,
        generateImage=Component.UNDEFINED,
        imageData=Component.UNDEFINED,
        responsive=Component.UNDEFINED,
        extent=Component.UNDEFINED,
        clearOnUnhover=Component.UNDEFINED,
        **kwargs
    ):
        self._prop_names = [
            "id",
            "autoRefreshLayout",
            "autolock",
            "autoungrabify",
            "autounselectify",
            "boxSelectionEnabled",
            "className",
            "clearOnUnhover",
            "contextMenu",
            "contextMenuData",
            "elements",
            "extent",
            "generateImage",
            "imageData",
            "layout",
            "maxZoom",
            "minZoom",
            "mouseoverEdgeData",
            "mouseoverNodeData",
            "pan",
            "panningEnabled",
            "responsive",
            "selectedEdgeData",
            "selectedNodeData",
            "style",
            "stylesheet",
            "tapEdge",
            "tapEdgeData",
            "tapNode",
            "tapNodeData",
            "userPanningEnabled",
            "userZoomingEnabled",
            "wheelSensitivity",
            "zoom",
            "zoomingEnabled",
        ]
        self._valid_wildcard_attributes = []
        self.available_properties = [
            "id",
            "autoRefreshLayout",
            "autolock",
            "autoungrabify",
            "autounselectify",
            "boxSelectionEnabled",
            "className",
            "clearOnUnhover",
            "contextMenu",
            "contextMenuData",
            "elements",
            "extent",
            "generateImage",
            "imageData",
            "layout",
            "maxZoom",
            "minZoom",
            "mouseoverEdgeData",
            "mouseoverNodeData",
            "pan",
            "panningEnabled",
            "responsive",
            "selectedEdgeData",
            "selectedNodeData",
            "style",
            "stylesheet",
            "tapEdge",
            "tapEdgeData",
            "tapNode",
            "tapNodeData",
            "userPanningEnabled",
            "userZoomingEnabled",
            "wheelSensitivity",
            "zoom",
            "zoomingEnabled",
        ]
        self.available_wildcard_properties = []
        _explicit_args = kwargs.pop("_explicit_args")
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs and excess named props
        args = {k: _locals[k] for k in _explicit_args}

        super(Cytoscape, self).__init__(**args)



import os
import importlib
import pytest
import time


def create_usage_test(dash_duo, filename, dir_name="usage"):
    app = importlib.import_module(filename).app

    dash_duo.start_server(app)
    dash_duo.wait_for_element_by_id("cytoscape", 20)

    # Wait for the flickr images to load
    if filename == "demos.usage-breadthfirst-layout":
        time.sleep(1)

    directory_path = os.path.join(os.path.dirname(__file__), "screenshots", dir_name)

    # Create directory if it doesn't already exist
    if not os.path.exists(directory_path):
        os.makedirs(directory_path)

    dash_duo.driver.save_screenshot(
        os.path.join(
            os.path.dirname(__file__), "screenshots", dir_name, filename + ".png"
        )
    )


@pytest.mark.parametrize(
    "name",
    [
        "usage-advanced",
        "demos.usage-animated-bfs",
        "demos.usage-breadthfirst-layout",
        "demos.usage-compound-nodes",
        "usage-events",
        "usage-elements",
        "demos.usage-pie-style",
        "usage",
        "usage-stylesheet",
        "demos.usage-initialisation",
        "demos.usage-linkout-example",
        "demos.usage-image-export",
        "demos.usage-responsive-graph",
    ],
)
def test_cyug001_usage(name, dash_duo):
    create_usage_test(dash_duo, name)



import io
import json
import os
from setuptools import setup, find_packages


with open(os.path.join("dash_cytoscape", "package.json")) as f:
    package = json.load(f)

package_name = package["name"].replace(" ", "_").replace("-", "_")

setup(
    name=package_name,
    version=package["version"],
    author=package["author"],
    author_email=package["author-email"],
    packages=find_packages(include=[package_name, package_name + ".*"]),
    include_package_data=True,
    license=package["license"],
    description=package["description"] if "description" in package else package_name,
    long_description=io.open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    url="https://dash.plotly.com/cytoscape",
    install_requires=[
        "dash",
    ],
    extras_require={
        "leaflet": ["dash-leaflet>=1.0.16rc3"],
    },
    python_requires=">=3.8",
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Flask",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Financial and Insurance Industry",
        "Intended Audience :: Healthcare Industry",
        "Intended Audience :: Manufacturing",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Database :: Front-Ends",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
    ],
)



import os
import importlib

from selenium.webdriver.common.keys import Keys
from selenium.webdriver.common.by import By


def create_input_and_save(
    dash_duo, css_selector, dir_name, options, prefix=None, name_map=None, save=True
):
    elem = dash_duo.find_element(css_selector)

    directory_path = os.path.join(os.path.dirname(__file__), "screenshots", dir_name)

    # Create directory if it doesn't already exist
    if not os.path.exists(directory_path):
        os.makedirs(directory_path)

    if prefix and not name_map:
        name_map = {option: prefix + option for option in options}

    elif not name_map:
        name_map = {}

    for option in options:
        elem.send_keys(Keys.CONTROL + "a")
        elem.send_keys(option)
        elem.send_keys(Keys.RETURN)

        if save:
            # If the name map doesn't contain a custom name for the option,
            # we default to the value of the option to name the saved
            # screenshot
            name = name_map.get(option, option)
            name = name.replace("(", "_").replace(")", "").replace(",", "_")
            name = name.replace("#", "_hex_").replace(" ", "")

            dash_duo.wait_for_element_by_id("cytoscape", 20)

            path = os.path.join(
                os.path.dirname(__file__), "screenshots", dir_name, name + ".png"
            )

            dash_duo.driver.save_screenshot(path)


def click_button_and_save(dash_duo, name_to_xpaths, dir_name, save=True):
    for name, xpath in name_to_xpaths.items():
        button = dash_duo.driver.find_element(By.XPATH, xpath)
        button.click()

        if save:
            dash_duo.wait_for_element_by_id("cytoscape", 20)

            path = os.path.join(
                os.path.dirname(__file__), "screenshots", dir_name, name + ".png"
            )

            dash_duo.driver.save_screenshot(path)


def test_cycb001_callbacks(dash_duo):
    app = importlib.import_module("usage-advanced").app

    dash_duo.start_server(app)
    dash_duo.wait_for_element_by_id("cytoscape", 20)

    create_input_and_save(
        dash_duo,
        css_selector="#dropdown-select-element-list input",
        dir_name="elements",
        options=["Basic", "Compound", "Gene", "Wineandcheese"],
    )

    create_input_and_save(
        dash_duo,
        css_selector="#dropdown-layout input",
        dir_name="layouts",
        options=["Preset", "Grid", "Circle", "Concentric", "Breadthfirst", "Cose"],
    )

    # Reset the input to what it was at the beginning
    create_input_and_save(
        dash_duo,
        css_selector="#dropdown-select-element-list input",
        dir_name="elements",
        options=["Basic"],
        save=False,
    )
    create_input_and_save(
        dash_duo,
        css_selector="#dropdown-layout input",
        dir_name="layouts",
        options=["Circle"],
        save=False,
    )

    # Input Different types of Node Content
    create_input_and_save(
        dash_duo,
        css_selector="#input-node-content",
        dir_name="style",
        options=["Hello", "data(id)"],
        name_map={"Hello": "NodeDisplayContentStatic", "data(id)": "NodeDisplayID"},
    )

    # Input Different node widths
    create_input_and_save(
        dash_duo,
        css_selector="#input-node-width",
        dir_name="style",
        options=["30", "50"],
        prefix="NodeWidth",
    )

    # Input Different node heights
    create_input_and_save(
        dash_duo,
        css_selector="#input-node-height",
        dir_name="style",
        options=["40", "60"],
        prefix="NodeHeight",
    )

    # Input different node shapes
    create_input_and_save(
        dash_duo,
        css_selector="#dropdown-node-shape input",
        dir_name="style",
        options=[
            "Triangle",
            "Rectangle",
            "Roundrectangle",
            "Barrel",
            "Diamond",
            "Pentagon",
            "Star",
            "Tag",
            "Vee",
            "Ellipse",
        ],
        prefix="NodeShape",
    )

    create_input_and_save(
        dash_duo,
        css_selector="#input-node-color",
        dir_name="style",
        options=["pink", "sky blue", "rgb(186,44,162)", "#def229"],
        prefix="NodeColor",
    )

    create_input_and_save(
        dash_duo,
        css_selector="#input-node-border-width",
        dir_name="style",
        options=["2"],
        save=False,
    )

    create_input_and_save(
        dash_duo,
        css_selector="#input-node-border-color",
        dir_name="style",
        options=["pink", "sky blue", "rgb(186,44,162)", "#def229"],
        prefix="BorderColor",
    )

    create_input_and_save(
        dash_duo,
        css_selector="#input-node-border-width",
        dir_name="style",
        options=["5", "2"],
        prefix="NodeBorderWidth",
    )

    create_input_and_save(
        dash_duo,
        css_selector="#dropdown-node-border-style input",
        dir_name="style",
        options=[
            "Dashed",
            "Dotted",
            "Double",
            "Solid",
        ],
        prefix="NodeBorderStyle",
    )

    create_input_and_save(
        dash_duo,
        css_selector="#input-node-padding",
        dir_name="style",
        options=["5px"],
        prefix="NodePadding",
    )

    create_input_and_save(
        dash_duo,
        css_selector="#dropdown-node-padding-relative-to input",
        dir_name="style",
        options=["Width", "Height", "Average", "Min", "Max"],
        prefix="NodePaddingRelativeTo",
    )

    create_input_and_save(
        dash_duo,
        css_selector="#input-edge-line-width",
        dir_name="style",
        options=["10", "1", "3"],
        prefix="LineWidth",
    )

    create_input_and_save(
        dash_duo,
        css_selector="#dropdown-edge-curve-style input",
        dir_name="style",
        options=["Haystack", "Segments", "Unbundled-bezier", "Bezier"],
        prefix="EdgeCurveStyle",
    )

    create_input_and_save(
        dash_duo,
        css_selector="#input-edge-line-color",
        dir_name="style",
        options=["pink", "sky blue", "rgb(186,44,162)", "#def229"],
        prefix="EdgeColor",
    )

    # Modify Edge Styles
    click_button_and_save(
        dash_duo,
        name_to_xpaths={
            "EdgeStyleSolid": '//*[@id="radio-edge-line-style"]/label[1]',
            "EdgeStyleDotted": '//*[@id="radio-edge-line-style"]/label[2]',
            "EdgeStyleDashed": '//*[@id="radio-edge-line-style"]/label[3]',
        },
        dir_name="style",
    )

    # Set "Use Edge Arrow" to "Yes"
    click_button_and_save(
        dash_duo,
        name_to_xpaths={"EdgeArrow": '//*[@id="radio-use-edge-arrow"]/label[1]'},
        dir_name="style",
        save=False,
    )

    create_input_and_save(
        dash_duo,
        css_selector="#dropdown-source-arrow-shape input",
        dir_name="style",
        options=["Circle", "Vee", "Tee", "Diamond", "Triangle"],
        prefix="EdgeArrowShape",
    )

    create_input_and_save(
        dash_duo,
        css_selector="#input-source-arrow-color",
        dir_name="style",
        options=["pink", "sky blue", "rgb(186,44,162)", "#def229"],
        prefix="EdgeArrowColor",
    )

    click_button_and_save(
        dash_duo,
        name_to_xpaths={
            "EdgeArrowFilled": ('//*[@id="radio-source-arrow-fill"]/' "label[1]"),
            "EdgeArrowHollow": ('//*[@id="radio-source-arrow-fill"]/' "label[2]"),
        },
        dir_name="style",
    )

    create_input_and_save(
        dash_duo,
        css_selector="#input-arrow-scale",
        dir_name="style",
        options=["3", "2", "1"],
        prefix="EdgeArrowScale",
    )


def test_cycb002_callbacks(dash_duo):
    app = importlib.import_module("usage-context-menu").app
    dash_duo.start_server(app)
    dash_duo.wait_for_element_by_id("cytoscape", 20)

    def test_context_menu_after_cb(dash_duo, options, css_selector="#dropdown input"):
        elem = dash_duo.find_element(css_selector)
        for option in options:
            elem.send_keys(Keys.CONTROL + "a")
            elem.send_keys(option)
            elem.send_keys(Keys.RETURN)
            css_selector = f"button#{option}"
            try:
                dash_duo.find_element(css_selector)
                assert True
            except Exception:
                assert False

    test_context_menu_after_cb(
        dash_duo,
        options=["add-node", "remove", "add-edge"],
    )
